#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <gpe/errorbox.h>
//#include <apm.h>

#include "main.h"
#include "sound.h"
#include "sms.h"
#include "phoneComms.h"
#include "term.h"
#include "pdu.h"
#include "db.h"
#include "io.h"


#ifndef PATH_MAX
  #define PATH_MAX 1024
#endif

GIOChannel *iochan;

int tty_fd;
int LogFile, bufIndex;
#define BUFFER_SIZE 1024
char at_buf[BUFFER_SIZE];
time_t CallStartTime, LastUpdateTime;
TPhoneState phoneState = psIdle;
TPhoneSecurity phoneSecurity = psUnknown;


time_t lastReadTime;
int bufEndIndex = -1;
int creg;
int cmut;
const char CRLFString[] = "\r\n";
gboolean gotOK, reading = FALSE;
gboolean got_reply=FALSE;
gboolean clck_ps=TRUE;

int debug=4;
int clcc_len;
int signal_v1;
int signal_v2;

gint phone_fun;
gint phone_cops;
gint in_command;
static char *pdu;
int sms_id;

int open_phone_connection_from_tty_interface()
{
	int	r;
	int	retVal;
	
	retVal = -1;
	tty_fd = open(phoneConfig.devName, O_RDWR | O_NONBLOCK);
	if (tty_fd >= 0) {
		r	= term_set(tty_fd,1,115200,P_NONE,8,FC_NONE,1,1);
		if (r >= 0) {
			r	= term_apply(tty_fd);
			if (r >= 0) {
				retVal	= r;
			}
		}
	}
	return retVal;
}

gboolean openPhone(void)
{
	int r;
	char *errmsg=NULL;

	LogFile = open("/tmp/gomunicator.log", O_WRONLY | O_CREAT | O_APPEND, 00666);
	r	= term_lib_init();
	if (r >= 0) {
		r	= open_phone_connection_from_tty_interface();
		if (r < 0) {
			errmsg	= g_strdup_printf("cannot open connection to phone module in %s: %s", 
					phoneConfig.devName,
					strerror(errno));
		}
	}
	else
	{
		errmsg	= g_strdup_printf("term_lib_init() failed: %s",  
					term_strerror(term_errno, errno));		
	}
	if (r >= 0)
	{
		return TRUE;
	}
	else
	{
		gpe_error_box(errmsg);
		g_free(errmsg);
		return FALSE;
	}
}

ssize_t
writen_ni(int fd, const void *buff, size_t n)
{
	size_t nl;
	ssize_t nw;
	const char *p;

	p = buff;
	nl = n;
	while (nl > 0) {
		do {
		       nw = write(fd, p, nl);
		} while ( nw < 0 && errno == EINTR );
		if ( nw <= 0 ) break;
		nl -= nw;
		p += nw;
	}

	return n - nl;
}

gboolean sendCommand(const char *cmd)
{
	const char prefix[2] = { 0x02, 0x16 };
	const char postfix[2] = { 0x16, 0x02 };
	if (debug >= 2)
		fprintf(stderr, "sendCommand('%s')\n", cmd);
	if (phoneConfig.at_cmds_in_frames == TRUE)
	{
		if (writen_ni(tty_fd, prefix, 2) < 2)
			return FALSE;
	}
	if (writen_ni(tty_fd, cmd, strlen(cmd)) < strlen(cmd))
		return FALSE;
	if (writen_ni(tty_fd, CRLFString, 2) < 2)
		return FALSE;
	if (phoneConfig.at_cmds_in_frames == TRUE)
	{
		if (writen_ni(tty_fd, postfix, 2) < 2)
			return FALSE;
	}
	in_command=1;
	return TRUE;
}

gboolean sendInitString(void)
{
#if 0
	time_t timeout;
	timeout = time(NULL) + 3;
	do {
		if (poll(&PhonePoll, 1, 0) == 1 &&
		    (PhonePoll.revents & ~(POLLIN | POLLPRI)) == 0)
			read(PhoneFile, at_buf, 1024);
	} while (time(NULL) < timeout);

	printf("sendInitString(), phoneInitString = %s", phoneConfig.phoneInitString);
	if (write(PhoneFile, phoneConfig.initATCmd, 12) != 12)
		return FALSE;

	timeout = time(NULL) + 3;
	do {
		if (poll(&PhonePoll, 1, 0) == 1 &&
		    (PhonePoll.revents & ~(POLLIN | POLLPRI)) == 0)
			read(PhoneFile, at_buf, 1024);
	} while (time(NULL) < timeout);
#endif
	return TRUE;
}				//sendInitString


gboolean phoneGetSecurity()
{
	if (tty_fd <= 0)
		return FALSE;

	if (!sendCommand("AT+CPIN?"))
		return FALSE;
	
	gotOK = FALSE;
	got_reply = FALSE;
	tryReadPhone(4);
	return gotOK && phoneSecurity != psUnknown;
}				//phoneGetSecurity



gboolean sendPIN(const gchar * PINString)
{
	sprintf(at_buf,"AT+CPIN=\"%s\"", PINString);
	return phoneGetOK(at_buf, TRUE);
}

/**
 * Sets all sorts of rubbish for the HTC Himalaya phone.
 * Valid parameter values are:
 * 0 = Set minimum functionality. Actually turn off the phone.
 * 1 = Make hardware reset and set full functionality.
 */
gboolean sendSetEquipmentState(int fun)
{
	int i = 0;
	/* V0 says "don't do "OK" do "0" instead */
	/*sprintf(at_buf,"ATZ;S0=0;E0;Q0;V0;X3;&C1");
	phoneGetOK(at_buf, TRUE);*/

	sprintf(at_buf,"ATZ;S0=0;E0;Q0;X3;&C1");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"Tx:AT+CMEE=1;+CRC=1;+CR=1;+CMOD=0;+CREG=1");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT+FCLASS=0;+CSSN=0,1;+CLIP=1;+COLP=0;+CCWA=1;+CMGF=0;+CSCS=\"HEX\"");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT+CGREG=1");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT+CUSD=1");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT%%SATC=1");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT+CMUT=0");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT$AU=100");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT+CLIR=0");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT%%CGPPP=1");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT%%CPRIM=\"GMM\",\"CONFIG MULTISLOT_CLASS=<10>\"");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT+CREG=1");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT%%IATT");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT+CFUN=?");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT%%SATC=1,FFFFFF7F00000010;+CFUN=%d", fun);
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"ATE0");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT+CLCK=?");
	phoneGetOK(at_buf, TRUE); /* processing here changes clck_ps */

	phoneGetSecurity();

	sprintf(at_buf,"AT");
	phoneGetOK(at_buf, TRUE);

	phoneGetSecurity();

	sprintf(at_buf,"AT%%ALS=0");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT+CPBR=?");
	phoneGetOK(at_buf, TRUE); 

	sprintf(at_buf,"AT$AD=100");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT+CREG=?");
	phoneGetOK(at_buf, TRUE);

	setCOPS(0);

	sprintf(at_buf,"AT+CFUN=?");
	phoneGetOK(at_buf, TRUE);

	sprintf(at_buf,"AT+CMUT=?");
	phoneGetOK(at_buf, TRUE);

	/* check ten times for CLCK "PS" response, after which
	 * it looks like it's okay to check for a PIN number - not before!
	 */
	for (i = 0; (i < 3) && clck_ps; i++)
	{
		sprintf(at_buf,"AT+CLCK=?");
		phoneGetOK(at_buf, TRUE); /* processing here changes clck_ps */
		phoneGetSecurity();
		sleep(1);
	}

	return gotOK;
}

/**
 * Sets the phones mobile station level functionality.
 * Valid parameter values are:
 * 0 = Set minimum functionality. Actually turn off the phone.
 * 1 = Make hardware reset and set full functionality.
 */
gboolean sendFUN(int fun)
{
	sprintf(at_buf,"AT+CFUN=%d", fun);
	return phoneGetOK(at_buf, TRUE);
}

/**
 * Ask phones mobile station level funtionality. Return values
 * 1 == Phone is in full functionality
 * 0 == Phone is turned off
 */
gint getFUN(void)
{
	phone_fun=-1;
	if (! phoneGetOK("AT+CFUN?", TRUE))
		return -1;
	return phone_fun;
}

/**
 * Ask current operator (PLMN)
 */
gint getCOPS(void)
{
	phone_cops=-1;
	if (! phoneGetOK("AT+COPS?", TRUE))
		return -1;
	return phone_cops;
}

/**
 * Select operator (PLMN)
 */
gboolean setCOPS(int cops)
{
	sprintf(at_buf,"AT+COPS=%d", cops);
	return phoneGetOK_slow(at_buf, TRUE);
}

gboolean sendSMS(char *sms_pdu)
{
	gboolean	retVal;
	
	pdu=sms_pdu;
	if (phoneConfig.wait_ack_before_sending_sms_pdu == FALSE)
	{
		if (debug >= 3)	
			printf("sendSMS(), do not wait ack before sending PDU, SMS PDU = %s\n", pdu);
		sprintf(at_buf,"AT+CMGS=%d%c%s%c", strlen(sms_pdu)/2 - 1, 0x0d, pdu, 0x1a);
	}
	else
	{
		if (debug >= 3)	
			printf("sendSMS(), wait ack before sending PDU, SMS PDU = %s, \n", pdu);
		sprintf(at_buf,"AT+CMGS=%d", strlen(sms_pdu)/2);
	}
	retVal	= phoneGetOK_slow(at_buf, TRUE);
	return retVal;	
}

gboolean phoneGetBattery(void)
{
	if (tty_fd <= 0 || reading)
		// return FALSE;
		return TRUE;
	if (write(tty_fd, "AT+CBC\r", 7) != 7)
		// return FALSE;
		return TRUE;
	return FALSE;
}				//phoneGetBattery


gboolean phoneGetSignal(void)
{
	if (tty_fd <= 0 || reading)
		return FALSE;

	if (write(tty_fd, "AT+CSQ\r", 7) != 7)
		return FALSE;
	return TRUE;
}				//phoneGetBattery

gboolean
ring_timeout(gpointer data)
{

	fprintf(stderr,"ring_timeout\n");
	if (phoneState != psRinging)
		return FALSE;
	if (! in_command) {
		if (! clcc_len) {
			got_reply=FALSE;
			display_set_number(NULL);
			display_set_name(NULL);
			db_call_update_list_duration(callid,CALL_LIST_MISSED);
			callid=-1;
			StopRinging(TRUE);
			audio_off();
			return FALSE;
		}
		got_reply=FALSE;
		clcc_len=0;
		sendCommand("AT+CLCC");
	}
	return TRUE;
}


void processReadLine(char *line)
{
	if (debug >= 2)
		fprintf(stderr, "processReadLine('%s',%d)\n", line, strlen(line));
	if (!strcmp(line, "OK")) {
		gotOK = TRUE;
		got_reply = TRUE;
		in_command=0;
		return;
	}
	else if (!strcmp(line, "ERROR")) {
		gotOK = FALSE;
		got_reply = TRUE;
		in_command=0;
		return;
	}
	else if (!strncmp(line, "ATZ;", 4)) {
		/* the first himalaya command gets echoed.
		 * by pretending it's a reply we can keep
		 * the state machine happy...
		 */
		gotOK = FALSE;
		got_reply = TRUE;
		in_command=0;
		return;
	}
	else if (!strncmp(line, "+CME ERROR:", 11)) {
		gotOK = FALSE;
		got_reply = TRUE;
		in_command=0;
		return;
	}
	else if (!strncmp(line, "AT+UN: ", 7)) { /* no idea what this is! */
		gotOK = FALSE;
		got_reply = TRUE;
		in_command=0;
		return;
	}
	else if (!strncmp(line, "+CSQ: ", 6)) {
		char *comma;
		gotOK = FALSE;
		got_reply = TRUE;
		in_command=0;
		line+=6;
		comma = strchr(line, ',');
		if (comma == NULL)
		{
			signal_v1 = -1;
			signal_v2 = -1;
			return;
		}

		*comma = 0;
		comma++;
		signal_v1=atoi(line);
		signal_v2=atoi(comma);
		UpdateSignal(signal_v1 / 7 + 1);
		return;
	}
	else if (!strncmp(line, "+CREG: ", 7)) {
		gotOK = FALSE;
		got_reply = TRUE;
		in_command=0;
		line+=7;
		creg=atoi(line);		
		return;
	}
	else if (!strncmp(line, "+CMUT: ", 7)) {
		gotOK = FALSE;
		got_reply = TRUE;
		in_command=0;
		line+=7;
		cmut=atoi(line);		
		return;
	}
	else if (!strncmp(line, "+CLCK:", 6)) {
		gotOK = FALSE;
		got_reply = TRUE;
		in_command=0;
		clck_ps = (strstr(line, "\"PS\"") != NULL) ? TRUE : FALSE;
		return;
	}
	else if (!strncmp(line, "+CMS ERROR:", 11)) {
		gotOK = FALSE;
		got_reply = TRUE;
		in_command=0;
		return;
	}
	else if (!strcmp(line, "RING") || !strcmp(line,"+CRING: VOICE"))  {
		if (phoneState == psIdle) {
			if (callid == -1)
				callid=db_call_new("",CALL_LIST_UNKNOWN);
			phoneState = psRinging;
			audio_on();
			StartRinging();
			clcc_len=1;
			g_timeout_add(250, ring_timeout, NULL);
		} else 
			vibra_on();
		return;
	}
	else if (!strcmp(line, "NO CARRIER"))  {
		if (phoneState == psRinging) {
			display_set_number(NULL);
			display_set_name(NULL);
			db_call_update_list_duration(callid,CALL_LIST_MISSED);
			callid=-1;
			StopRinging(TRUE);
		}
		if (phoneState == psOnline) {
			db_call_update_duration(callid);
			callid=-1;
			EndCall();
		}
		audio_off();
		phoneState = psIdle;
		return;
	}
	else if (!strncmp(line, "+CPIN: ", 7)) {
		line+=7;
		if (!strcmp(line, "READY"))
		{
			gotOK = TRUE;
			phoneSecurity = psReady;
		}
		else if (!strcmp(line, "SIM PIN"))
			phoneSecurity = psNeedSIM_PIN;
		else
			phoneSecurity = psUnknown;		
		return;
	}
	else if (!strncmp(line, "+CFUN: ", 7)) {
		line+=7;
		phone_fun=atoi(line);		
		return;
	}
	else if (!strncmp(line, "+COPS: ", 7)) {
		char *end;
		line+=7;
		if (line[1] == ',') {
			phone_cops=1;
			line+=2;
			if (line[1] == ',') {
				line+=2;
				if (line[0]=='"') {
					line++;
					end=index(line,'"');
					if (end) {
						*end='\0';
					}
					display_set_operator(line);
				}	
			}
		} else
			phone_cops=0;
		return;
	}
	else if (!strncmp(line, "+CLIP: ", 7)) {
		char str[64],*end;
		line+=7;
		if (line[0]=='"') {
			line++;
			end=index(line,'"');
			if (end) {
				*end='\0';
			}
			sprintf(str,"Call from %s", line);
			if (callid == -1)
				callid=db_call_new(line,CALL_LIST_UNKNOWN);
			else
				db_call_update_number(callid,line);
			display_set_number(str);
			display_lookup_set_name(line);
		}
		return;
	}
	else if (!strncmp(line, "+CMTI: \"SM\",", 12)) {
		line+=12;
		NewSMS(atoi(line));
		sendCommand("AT+CMGL=0");
		return;
	}
	else if (!strncmp(line, "+CLCC: ", 7)) {
		clcc_len++;
		return;
	}
	else if (!strncmp(line, "+CMGS: ", 7)) {
		sms_id=atoi(line+7);
		return;
	}
	else if (!strncmp(line, "+CPMS: ", 7)) {
		printf("%s\r\n", line);
		return;
	}	
	else if (line[0] >= '0' && line[0] <= '9') {
		char filename[PATH_MAX];		
		struct pdu *p;
		FILE *f;
		
		sprintf(filename,"%s/.sms",getenv("HOME"));
		f=fopen(filename,"a");
		if (f) {
			fprintf(f, "%s\n", line);
			fclose(f);
		}
		db_sms_new("Inbox",line);
		p=pdu_decode(line);
		if (p->text) 
			newSMS(p->text);	
		pdu_free(p);	
		return;
	}
	else if (line[0] == '>' && pdu) {
		if (phoneConfig.wait_ack_before_sending_sms_pdu == TRUE) {
			sendCommand(pdu);
			phoneGetOK_slow("\032", TRUE);
		}
		pdu = NULL;
		return;
	}
	else if (debug >= 1)
		fprintf(stderr,"Unknown Reply '%s'\n", line);
#if 0
	if (!strcmp(line, "+CBC: ", 5)) {
		switch (at_buf[bufIndex]) {
				case '1':
					charging = TRUE;
				case '0':
					counter = 0;
					for (bufIndex += 2;
					     at_buf[bufIndex] >= '0';
					     ++bufIndex)
						counter =
						    counter * 10 +
						    at_buf[bufIndex] -
						    '0';
					UpdateBattery(counter / 19,
						      charging);

					/* add stat function here ? */
					int va, vb, vc;
//  apm_info info;                            
					//apm_read(&info);                          

					// va = info.battery_time;    
					//vb = info.battery_percenta
					// v//c = info.ac_line_status;  
//  UpdateBattery(vb / 19, charging);      
					break;
				default:
					UpdateBattery(-1, FALSE);
					break;
				}	//switch (at_buf[bufIndex])

			} else
			    if (strncmp(at_buf + bufIndex, "CSQ: ", 5)
				== 0) {
				counter = 0;
				for (bufIndex += 5;
				     at_buf[bufIndex] >= '0';
				     ++bufIndex)
					counter =
					    counter * 10 +
					    at_buf[bufIndex] - '0';
				UpdateSignal(counter / 7 + 1);

			} else
			    if (strncmp
				(at_buf + bufIndex, "CMTI: \"SM\", ",
				 12)
				== 0) {
				bufIndex += 12;
				for (counter = 0;
				     at_buf[bufIndex] >= '0';
				     ++bufIndex)
					counter =
					    counter * 10 +
					    at_buf[bufIndex] - '0';
				NewSMS(counter);

/*        } else if (strncmp(at_buf + bufIndex, "CMGR: ", 6) == 0) {
          bufIndex += 6;
          processMessage(at_buf + bufIndex);
*/
			} else
			    if (strncmp
				(at_buf + bufIndex, "CPMS: ",
				 6) == 0) {
				printf("%s\r\n", at_buf + bufIndex);
				bufIndex += 6;
				for (counter = 0;
				     at_buf[bufIndex] >= '0'
				     && at_buf[bufIndex] <= '9';
				     ++bufIndex)
					counter =
					    counter * 10 +
					    at_buf[bufIndex] - '0';
				printf("NewSMS(%d)\r\n", counter);
				NewSMS(counter);
				while (at_buf[bufIndex] >= ',')
					++bufIndex;

			} else
			}	//if (strncmp(at_buf + bufIndex, "...) == 0)...else if...
		}		//switch (at_buf[bufIndex])
		while (bufIndex < lineEndIndex &&
		       strncmp(at_buf + bufIndex, CRLFString, 2) != 0)
			++bufIndex;
		while (bufIndex <= lineEndIndex
		       && at_buf[bufIndex] <= ' ')
			++bufIndex;
	} while (bufIndex < lineEndIndex);
//  lineEndIndex = -1;
#endif
}				//processReadLine

int
phone_strip_garbage(int *end)
{
	int pos=0,removed=0;
	char *src=at_buf;
	char *dst=at_buf;
	while (pos <= *end) {
		if (*src >= ' ' || *src == '\r' || *src == '\n' || *src == '\0')
			*dst++=*src;
		else
			removed++;
		src++;
		pos++;
	}
	*dst='\0';
	*end-=removed;
	return removed;
}

int
phone_check_complete(int *end)
{
	int pos=0;
	while (pos <= *end) {
		if (at_buf[pos] == '\r' || at_buf[pos] == '\n' || (at_buf[0] == '>' && pos > 0)) {
			char line[pos+1];
			memcpy(line, at_buf, pos);
			line[pos++]='\0';
			memcpy(at_buf, at_buf+pos, BUFFER_SIZE-pos);
			*end-=pos;
			if (line[0])
				processReadLine(line);
			return pos;
		}
		pos++;
	}
	return 0;
}
void
dump_buffer(char *text, char *str, int len)
{
	fprintf(stderr,"%s",text);
	while (len > 0) {
		if (*str >= ' ')
			fprintf(stderr,"%c", *str++);
		else
			fprintf(stderr,"(%02x)", *str++);
		len--;
	}
}

void ReadPhoneStartGarbage(int timeout)
{
	int readLength, sel_res, remove;
	fd_set sel_readfds;
	struct timeval sel_timeout;
	char data[1024];

	if (debug >= 3)
		fprintf(stderr, "ReadPhoneStartGarbage(%d)\n", timeout);

	if (tty_fd <= 0)
		return;
	
	FD_ZERO(&sel_readfds);
	sel_timeout.tv_sec=timeout;
	sel_timeout.tv_usec=0;

	FD_SET(tty_fd, &sel_readfds);
	sel_res=select(tty_fd+1, &sel_readfds, NULL, NULL, &sel_timeout);
	if (debug >= 4)
		fprintf(stderr,"sel_res=%d\n", sel_res);
	if (! sel_res) {
		if (debug >= 3)
			fprintf(stderr, "ReadphoneStartGarbage return, timeout\n");
		return;
	}
			
	readLength = read(tty_fd, data, sizeof(data));
	if (debug >= 4)
		fprintf(stderr,"readLength %d\n", readLength);
	if (readLength > 0) {
		if (debug >= 4) {
			dump_buffer("Buffer is ",at_buf,readLength);
		}
	}		//if (readLength != 0)

	if (debug >= 3)
		fprintf(stderr, "Readphone return\n");
}				//tryReadPhone

gboolean
phone_data_available(GIOChannel *iochan, GIOCondition condition, gpointer t)
{
	if (condition == G_IO_IN) {
		tryReadPhone(0);
		return TRUE;
	}
	return FALSE;
}

void tryReadPhone(int timeout)
{
	int readLength, sel_res, remove;
	fd_set sel_readfds;
	struct timeval sel_timeout;

	if (debug >= 3)
		fprintf(stderr, "tryReadPhone(%d)\n", timeout);

	if (tty_fd <= 0)
		return;
	
	FD_ZERO(&sel_readfds);
	sel_timeout.tv_sec=timeout;
	sel_timeout.tv_usec=0;
	while (bufEndIndex < 1023 && (!got_reply || !timeout))
	{
		FD_SET(tty_fd, &sel_readfds);
		sel_res=select(tty_fd+1, &sel_readfds, NULL, NULL, &sel_timeout);
		if (debug >= 4)
			fprintf(stderr,"sel_res=%d\n", sel_res);
		if (! sel_res) {
			if (debug >= 3)
				fprintf(stderr, "tryReadphone return, timeout\n");
			return;
		}
				
		readLength = read(tty_fd, at_buf+bufEndIndex+1, 1023 - bufEndIndex);
		if (debug >= 4)
			fprintf(stderr,"readLength %d\n", readLength);
		if (readLength > 0) {
			if (debug >= 4)
				fprintf(stderr,"bufEndIndex=%d\n", bufEndIndex);
			bufEndIndex+=readLength;
			do {
				if (debug >= 4) {
					dump_buffer("Buffer is ",at_buf,bufEndIndex+1);
					fprintf(stderr," length=%d bufEndIndex=%d\n", strlen(at_buf), bufEndIndex);
				}
				remove=phone_strip_garbage(&bufEndIndex);
				if (debug >= 4)
					fprintf(stderr,"strip=%d\n", remove);
				remove=phone_check_complete(&bufEndIndex);
				if (debug >= 4)
					fprintf(stderr,"complete=%d\n", remove);
			} while (remove);
		}		//if (readLength != 0)
	}

	if (debug >= 3)
		fprintf(stderr, "tryReadphone return\n");
}				//tryReadPhone

void
phone_io_setup(void)
{
	got_reply=FALSE;
	gotOK=FALSE;
	iochan=g_io_channel_unix_new(tty_fd);
	g_io_add_watch(iochan, G_IO_IN|G_IO_ERR|G_IO_HUP, phone_data_available, NULL);
}

gboolean phoneGetOK(const char *query, gboolean waitForRead)
{
	if (tty_fd <= 0)
		return FALSE;

	if (!sendCommand(query))
		return FALSE;

	gotOK = FALSE;
	got_reply = FALSE;
	tryReadPhone(3);

	return gotOK;
}				//phoneGetOK

gboolean phoneGetOK_slow(const char *query, gboolean waitForRead)
{
	if (tty_fd <= 0)
		return FALSE;

	if (!sendCommand(query))
		return FALSE;

	gotOK = FALSE;
	got_reply = FALSE;
	tryReadPhone(20);

	return gotOK;
}				//phoneGetOK


char *scanPhoneBuffer(const char *key, int start)
{
	int keyLength = strlen(key);

	while (start <= bufEndIndex &&
	       strncmp(key, at_buf + start, keyLength) != 0)
		++start;
	if (start > bufEndIndex)
		return NULL;
	return at_buf + start + strlen(key);
}				//scanPhoneBuffer


void answerPhone(void)
{
	if (phoneState == psRinging) {
		if (phoneGetOK("ATA", TRUE)) {
			phoneState = psOnline;
			CallStartTime = LastUpdateTime = time(NULL);
			db_call_update_list_time(callid,CALL_LIST_INCOMING);
		} else {
			db_call_update_list_duration(callid,CALL_LIST_MISSED);
			callid=-1;
			StopRinging(FALSE);
			hangupPhone();
		}		//if (phoneGetReply("ATA"))...else
	}			//if (phoneState == psRinging)
}				//answerPhone


void hangupPhone(void)
{
	int stat;
	if (phoneState != psIdle) {
		audio_off();
		stat=phoneState;
		phoneState = psIdle;
		phoneGetOK("ATH0", FALSE);
		if (stat == psRinging)
			db_call_update_list_duration(callid,CALL_LIST_REJECTED);
		else
			db_call_update_duration(callid);
		callid=-1;
		display_set_number(NULL);
		display_set_name(NULL);
		tryReadPhone(2);
		phoneState = psIdle;
	}			//if (phoneState != psIdle)
}				//hangupPhone


gboolean makeCall(char *dialString)
{
	if (phoneState != psIdle)
		return FALSE;
	audio_on();
	if (!phoneGetOK(dialString, TRUE))
		return FALSE;
	phoneState = psOnline;
	CallStartTime = LastUpdateTime = time(NULL);
	return TRUE;
}				//makeCall


void closePhone(void)
{
	hangupPhone();
	if (tty_fd > 0) {
		close(tty_fd);
		tty_fd = 0;
	}			//if (PhoneFile > 0)
	if (LogFile > 0) {
		close(LogFile);
		LogFile = 0;
	}			//if (LogFile > 0)
}				//closePhone
